! ***********************************************************************
!
!   Copyright (C) 2010  Bill Paxton
!
!   this file is part of mesa.
!
!   mesa is free software; you can redistribute it and/or modify
!   it under the terms of the gnu general library public license as published
!   by the free software foundation; either version 2 of the license, or
!   (at your option) any later version.
!
!   mesa is distributed in the hope that it will be useful, 
!   but without any warranty; without even the implied warranty of
!   merchantability or fitness for a particular purpose.  see the
!   gnu library general public license for more details.
!
!   you should have received a copy of the gnu library general public license
!   along with this software; if not, write to the free software
!   foundation, inc., 59 temple place, suite 330, boston, ma 02111-1307 usa
!
! ***********************************************************************
 
      module run_star_extras

      use star_lib
      use star_def
      use const_def
      
      implicit none
      
      ! these routines are called by the standard run_star check_model
      contains

      subroutine extras_controls(s, ierr)
         type (star_info), pointer :: s
         integer, intent(out) :: ierr
         ierr = 0
         
         ! this is the place to set any procedure pointers you want to change
         ! e.g., other_wind, other_mixing, other_energy  (see star_data.inc)
         
         
      end subroutine extras_controls
      
      
      integer function extras_startup(s, id, restart, ierr)
         use run_star_support, only: change_lnPgas_flag, new_lnPgas_flag
         type (star_info), pointer :: s
         integer, intent(in) :: id
         logical, intent(in) :: restart
         integer, intent(out) :: ierr
         ierr = 0
         extras_startup = 0
         if (.not. restart) then
            call alloc_extra_info(s)
         else ! it is a restart
            call unpack_extra_info(s)
         end if
         
         write(*,*) change_lnPgas_flag, new_lnPgas_flag
         
      end function extras_startup
      

      ! returns either keep_going, retry, backup, or terminate.
      integer function extras_check_model(s, id, id_extra)
         type (star_info), pointer :: s
         integer, intent(in) :: id, id_extra
         extras_check_model = keep_going         
         if (.false. .and. s% star_mass_h1 < 0.35d0) then
            ! stop when star hydrogen mass drops to specified level
            extras_check_model = terminate
            write(*, *) 'have reached desired hydrogen mass'
            return
         end if
         
         s% xa_central_lower_limit_species(1) = 'h1' ! isotope name as defined in chem_def
         s% xa_central_lower_limit(1) = 0.6

         s% overshoot_f_above_nonburn = 0.01
         s% overshoot_f_below_nonburn = 0.01
         s% overshoot_f_above_burn_h = 0.01
         s% overshoot_f_below_burn_h = 0.01

         ! if you want to check multiple conditions, it can be useful
         ! to set a different termination code depenending on which
         ! condition was triggered.  MESA provides 9 customizeable
         ! termination codes, named t_xtra1 .. t_xtra9.  You can
         ! customize the messages that will be printed upon exit by
         ! setting the corresponding termination_code_str value.
         ! termination_code_str(t_xtra1) = 'my termination conditon'

         ! by default, indicate where (in the code) MESA terminated
         if (extras_check_model == terminate) s% termination_code = t_extras_check_model
      end function extras_check_model


      integer function how_many_extra_history_columns(s, id, id_extra)
         type (star_info), pointer :: s
         integer, intent(in) :: id, id_extra
         how_many_extra_history_columns = 0
      end function how_many_extra_history_columns
      
      
      subroutine data_for_extra_history_columns(s, id, id_extra, n, names, vals, ierr)
         type (star_info), pointer :: s
         integer, intent(in) :: id, id_extra, n
         character (len=maxlen_history_column_name) :: names(n)
         real(dp) :: vals(n)
         integer, intent(out) :: ierr
         
         !note: do NOT add these names to history_columns.list
         ! the history_columns.list is only for the built-in log column options.
         ! it must not include the new column names you are adding here.
         
         ierr = 0
      end subroutine data_for_extra_history_columns

      
      integer function how_many_extra_profile_columns(s, id, id_extra)
         type (star_info), pointer :: s
         integer, intent(in) :: id, id_extra
         how_many_extra_profile_columns = 37
      end function how_many_extra_profile_columns
      
      
      subroutine data_for_extra_profile_columns(s, id, id_extra, n, nz, names, vals, ierr)
         use star_lib, only: star_get_gyre_info
         use graco_support, only: get_extra_graco_columns
         use num_lib, only: safe_log10
         type (star_info), pointer :: s
         integer, intent(in) :: id, id_extra, n, nz
         character (len=maxlen_profile_column_name) :: names(n)
         real(dp) :: vals(nz,n)
         integer, intent(out) :: ierr
         integer :: k
         
         integer :: j, n_atm
         real(dp) :: G, M_star, R_star, L_star
         logical, parameter :: add_center_point = .false.
         logical :: keep_surface_point, add_atmosphere
         real(dp), pointer, dimension(:) :: r, w, L, p, rho, T, N2, Gamma_1, nabla_ad, delta, nabla,  &
                                   kappa, kappa_rho, kappa_T, epsilon, epsilon_rho, epsilon_T, omega
         real(dp), pointer, dimension(:) :: lnq, cp, ln_free_e, aast, beta, gradr, h1, h2, he3, he4, &
                                   li7, be7, c12, c13, n14, n15, o16, o17, be9, si28
         
         ierr = 0
         
         keep_surface_point = s% keep_surface_point_for_pulse_info
         if (nz .eq. size(w, dim=1)) then
            add_atmosphere = s% add_atmosphere_to_pulse_info
         else
            add_atmosphere = .false.
         endif
         
           call star_get_gyre_info(s% id, keep_surface_point, add_atmosphere, &
              G, M_star, R_star, L_star, r, w, L, p, rho, T, &
              N2, Gamma_1, nabla_ad, delta, nabla,  &
              kappa, kappa_rho, kappa_T, &
              epsilon, epsilon_rho, epsilon_T, omega, ierr)
           
           call get_extra_graco_columns(id, add_center_point, keep_surface_point, &
                 add_atmosphere, lnq, cp, ln_free_e, aast, beta, gradr, h1, h2, he3, he4, li7, &
                 be7, c12, c13, n14, n15, o16, o17, be9, si28, ierr)

           
  
           names(1) = 'q_div_xq'
           names(2) = 'chiT_div_chiRho'
           names(3) = 'kappa'
           names(4) = 'kappa_rho'
           names(5) = 'kappa_T'
           names(6) = 'epsilon'
           names(7) = 'epsilon_rho'
           names(8) = 'epsilon_T'
           names(9) = 'omega'
           names(10) = 'pressure'
           names(11) = 'density'
           names(12) = 'temperature'
           names(13) = 'luminosity'
           names(14) = 'gamma1'
           names(15) = 'gradT'
           names(16) = 'grada'
           names(17) = 'brunt_N2'
           names(18) = 'lnq'
           names(19) = 'Cp'
           names(20) = 'ln_free_e'
           names(21) = 'A2'
           names(22) = 'Pg_div_P'
           names(23) = 'gradr'
           names(24) = 'h1'
           names(25) = 'h2'
           names(26) = 'he3'
           names(27) = 'he4'
           names(28) = 'li7'
           names(29) = 'be7'
           names(30) = 'c12'
           names(31) = 'c13'
           names(32) = 'n14'
           names(33) = 'n15'
           names(34) = 'o16'
           names(35) = 'o17'
           names(36) = 'be9'
           names(37) = 'si28'
           
!            write(*,*) size(w, dim=1), nz, s% nz, s% atm_structure_num_pts
           n_atm = 0
           if (add_atmosphere .eqv. .true.) n_atm = s% atm_structure_num_pts
           
           ! Extra Columns to Support GYRE Models         
           ! From Surface towards the Core
           do j = nz + n_atm, 1, -1
             k          = nz + n_atm + 1 - j
             vals(k,1)  = w(j)
             vals(k,2)  = delta(j)
             vals(k,3)  = kappa(j)
             vals(k,4)  = kappa_rho(j)
             vals(k,5)  = kappa_T(j)
             vals(k,6)  = epsilon(j)
             vals(k,7)  = epsilon_rho(j)
             vals(k,8)  = epsilon_T(j)
             vals(k,9)  = omega(j)
             vals(k,10) = p(j)
             vals(k,11) = rho(j)
             vals(k,12) = T(j)
             vals(k,13) = L(j)
             vals(k,14) = Gamma_1(j)
             vals(k,15) = nabla(j)
             vals(k,16) = nabla_ad(j)
             vals(k,17) = N2(j)
           enddo

           ! Extra Columns to Support GraCo Models         
           ! From Surface towards the Core
           do k = 1, nz
             j          = nz + 1 - k
             vals(j,18) = lnq(k)
             vals(j,19) = cp(k) 
             vals(j,20) = ln_free_e(k)
             vals(j,21) = aast(k)
             vals(j,22) = beta(k)
             vals(j,23) = gradr(k)
             vals(j,24) = h1(k)
             vals(j,25) = h2(k)
             vals(j,26) = he3(k)
             vals(j,27) = he4(k)
             vals(j,28) = li7(k)
             vals(j,29) = be7(k)
             vals(j,30) = c12(k)
             vals(j,31) = c13(k)
             vals(j,32) = n14(k)
             vals(j,33) = n15(k)
             vals(j,34) = o16(k)
             vals(j,35) = o17(k)
             vals(j,36) = be9(k)
             vals(j,37) = si28(k)
           enddo
         
         deallocate(r, w, L, p, rho, T, N2, Gamma_1, nabla_ad, delta, nabla,  &
                    kappa, kappa_rho, kappa_T, epsilon, epsilon_rho, epsilon_T, omega)
         deallocate(lnq, cp, ln_free_e, aast, beta, gradr, h1, h2, he3, he4, li7, be7, &
                    c12, c13, n14, n15, o16, o17, be9, si28)
         
      end subroutine data_for_extra_profile_columns
      

      ! returns either keep_going or terminate.
      ! note: cannot request retry or backup; extras_check_model can do that.
      integer function extras_finish_step(s, id, id_extra)
         type (star_info), pointer :: s
         integer, intent(in) :: id, id_extra
         integer :: ierr
         extras_finish_step = keep_going
         call store_extra_info(s)

         ! to save a profile, 
            ! s% need_to_save_profiles_now = .true.
         ! to update the star log,
            ! s% need_to_update_history_now = .true.

         ! see extras_check_model for information about custom termination codes
         ! by default, indicate where (in the code) MESA terminated
         if (extras_finish_step == terminate) s% termination_code = t_extras_finish_step
      end function extras_finish_step
      
      
      subroutine extras_after_evolve(s, id, id_extra, ierr)
         type (star_info), pointer :: s
         integer, intent(in) :: id, id_extra
         integer, intent(out) :: ierr
         ierr = 0
      end subroutine extras_after_evolve
      
      
      ! routines for saving and restoring extra data so can do restarts
         
         ! put these defs at the top and delete from the following routines
         !integer, parameter :: extra_info_alloc = 1
         !integer, parameter :: extra_info_get = 2
         !integer, parameter :: extra_info_put = 3
      
      
      subroutine alloc_extra_info(s)
         integer, parameter :: extra_info_alloc = 1
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_alloc)
      end subroutine alloc_extra_info
      
      
      subroutine unpack_extra_info(s)
         integer, parameter :: extra_info_get = 2
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_get)
      end subroutine unpack_extra_info
      
      
      subroutine store_extra_info(s)
         integer, parameter :: extra_info_put = 3
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_put)
      end subroutine store_extra_info
      
      
      subroutine move_extra_info(s,op)
         integer, parameter :: extra_info_alloc = 1
         integer, parameter :: extra_info_get = 2
         integer, parameter :: extra_info_put = 3
         type (star_info), pointer :: s
         integer, intent(in) :: op
         
         integer :: i, j, num_ints, num_dbls, ierr
         
         i = 0
         ! call move_int or move_flg    
         num_ints = i
         
         i = 0
         ! call move_dbl       
         
         num_dbls = i
         
         if (op /= extra_info_alloc) return
         if (num_ints == 0 .and. num_dbls == 0) return
         
         ierr = 0
         call star_alloc_extras(s% id, num_ints, num_dbls, ierr)
         if (ierr /= 0) then
            write(*,*) 'failed in star_alloc_extras'
            write(*,*) 'alloc_extras num_ints', num_ints
            write(*,*) 'alloc_extras num_dbls', num_dbls
            stop 1
         end if
         
         contains
         
         subroutine move_dbl(dbl)
            real(dp) :: dbl
            i = i+1
            select case (op)
            case (extra_info_get)
               dbl = s% extra_work(i)
            case (extra_info_put)
               s% extra_work(i) = dbl
            end select
         end subroutine move_dbl
         
         subroutine move_int(int)
            integer :: int
            i = i+1
            select case (op)
            case (extra_info_get)
               int = s% extra_iwork(i)
            case (extra_info_put)
               s% extra_iwork(i) = int
            end select
         end subroutine move_int
         
         subroutine move_flg(flg)
            logical :: flg
            i = i+1
            select case (op)
            case (extra_info_get)
               flg = (s% extra_iwork(i) /= 0)
            case (extra_info_put)
               if (flg) then
                  s% extra_iwork(i) = 1
               else
                  s% extra_iwork(i) = 0
               end if
            end select
         end subroutine move_flg
      
      end subroutine move_extra_info

      end module run_star_extras
      
